;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - hierarchy-widget.ss                                            #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;;
;; This is the hierarchy-widget file.
;; We derive a canvas, and then code the whole widget from scratch :)
;;
 
;; We declare the class...
(define hierarchy-widget%
  ;; We derive it...
  (class canvas%
    ;; Fields
    (define data '())
    ;; Why using a bitmap and a drawing context as a class field instead
    ;; of directly declare it in the on-paint body? The matter is,
    ;; that if we did so a bitmap would be instantiated for each call of
    ;; the on-paint method. After testing it, even on a pretty powerful
    ;; pc, if the on-paint method is called too often, the program runs out
    ;; of resources and returns an error when instantiating a bitmap after
    ;; a short while.
    (define bitmap #f)
    (define dc #f)
    (define x-offset 0)
    (define y-offset 0)
    
    ;; Constants
    (define k-left-margin 21)
    (define k-top-margin 2)
    (define k-line-height 17)
    (define k-horizontal-scroll-step 3)
    
    ;; Those constants are used for drawing
    (define k-white         (make-object color% 255 255 255))
    (define k-black         (make-object color% 0 0 0))
    (define k-gray          (make-object color% 160 160 160))
    (define k-white-brush   (make-object brush% k-white 'solid))
    (define k-active-brush  (make-object brush% k-black 'hilite))
    ;; We use the default panel background (which depends on the platform, windowing system, etc.)
    (define k-gray-brush    (make-object brush% (get-panel-background) 'solid))
    ;; Opaque is needed here
    (define k-dkgray-brush  (make-object brush% (make-object color% 132 132 132) 'opaque))
    (define k-invisible-pen (make-object pen%   (make-object color% 0 0 0) 0 'transparent))
    (define k-edit-font     (make-object font%  12 "Courier" 'default 'normal 'normal #f 
                                                'unsmoothed #t))
    
    ;; To be platform compliant, we avoid smoothed text under windows, because highlighted smoothed
    ;; text under windows does not look really nice.
    (define k-normal-font        (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'normal 'normal #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'normal 'normal #f 'smoothed #t)))
    
    (define k-italic-font        (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'italic 'normal #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'italic 'normal #f 'smoothed #t)))
    
    (define k-italic-smoothed-font (make-object font% 11 'system 'italic 'normal #f 'smoothed #t))
    
    (define k-bold-font          (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'normal 'bold #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'normal 'bold #f 'smoothed #t)))
    
    ;; Loading the bitmaps now will allow us to reduce time needed to draw the widget.
    (define k-less-bitmap (make-object bitmap% (build-path "images" "less.png") 'png #f))
    (define k-plus-bitmap (make-object bitmap% (build-path "images" "plus.png") 'png #f))
    (define k-project-mask-bitmap (make-object bitmap% 
                                               (build-path "images" "icons" "16x16" "project-mask.png")
                                               'png #f))
    ;; Why using a a-list instead of declaring each icon? Of course declaring each icon would be a little faster,
    ;; but by using a a-list we can add as many icons as needed without changing the code.
    (define k-icon-list (map (lambda (elt) (cons (string->symbol (substring elt 0 (- (string-length elt) 4))) 
                                                 (make-object bitmap% 
                                                   (build-path "images" "icons" "16x16" elt)
                                                   'png #f)))
                             ;; We look for the icons in the right directory...
                             (map (lambda (p) (path->string p))
                                  (directory-list (path->string (build-path "images" "icons" "16x16"))))
                             ))
    
    ;; Public methods
    (public set-data
            get-selected-element
            get-selected-element-id
            delete-selected
            update-in
            update-buttons
            swap-up-selected
            swap-down-selected
            swap-up-one-selected
            swap-down-one-selected
    )
    
    ;; The 'data' field modifier...
    (define (set-data new-data) 
      ;; We just update the data...
      (set! data new-data))
    
    ;; This function updates the local environment with the contents of the global environment...
    (define (update-in . L)
      ;; ...we first update the data list...
      (set-data project-data-global)
      ;; ...then the scroll bars...
      (update-scroll-bars)
      ;; ...the bitmap...
      (update-bitmap)
      ;; ...the buttons...
      (update-buttons)
      ;; ...and finally repaint the whole widget.
      (on-paint))
    
    ;; This function returns the "real" width of the widget, i.e.
    ;; the total width needed to display the whole widget's contents.
    (define (get-real-width)
      ;; This function will parse the whole list in order to find the line
      ;; which needs the more pixel (in width) to be drawn...
      (define (parse-data-list data-list offset)
        ;; The maximum has been saved into previous-max. Therefore we just need to return
        ;; previous-max if the list parsing has ended (that is, the list is empty).
        (if (null? data-list)
            0 
            ;; Otherwise, the element we are working with is the car of the list.
            (let ((current-element (car data-list)))
              ;; Then we are going to simulate the drawing in order to find the line width (in pixels).
              (let-values (((label-width label-height label-distance label-extra)
                             (send dc get-text-extent
                                   ;; We find the label or the substitution text if the widget has no label...
                                   (if (assq 'label (cdr current-element))
                                       (let ((label (cadr (assq 'label (cdr current-element)))))
                                         (if (and label (not (string=? label ""))) label "(no label)"))
                                       (string-append "<" (symbol->string (car current-element)) ">"))
                                   ;; And find the right font...
                                   (cond ((memq (caar data-list) '(frame project)) k-bold-font)
                                         ((container? (car data-list)) k-italic-font)
                                         (else k-normal-font))))
                           ;; Ok, we found the label width - we need to check the id width, now.
                          ((id-width id-height id-distance id-extra)
                           (send dc get-text-extent  (let ((id-v (assq 'id (cdar data-list))))
                                                       (if id-v
                                                           (string-append " [" (symbol->string (cadr (assq 'id (cdr current-element)))) "]") ""))
                                 ;; The id is drawn in an italic font...
                                 k-italic-font)))
          ;; We are now able to find the length of the current line, and therefore save it...
          (let* ((current-length (+ (* 20 offset) k-left-margin 20 label-width id-width 10)))
            ;; Do the current element have children?
            (if (and (assq 'children (cdr current-element))
                     (not (collapsed? current-element)))
                ;; If it has, consider the maximum line length of children, in the maximum evaluation...
                (max (parse-data-list (cdr (assq 'children (cdr current-element))) (add1 offset))
                     (parse-data-list (cdr data-list) offset)
                     current-length)
                ;; Otherwise, only consider the maximum line length of brothers (and of course the current line length).
                (max (parse-data-list (cdr data-list) offset)
                     current-length)))))))
      
      ;; And then, we call the function we just declared...
      (inexact->exact (parse-data-list data 0)))
    
    ;; This function returns the "real" height of the widget, i.e.
    ;; the total height needed to display the whole widget's contents.
    (define (get-real-height)
      ;; We add 1 to ensure the height is != 0...
      (add1 (* k-line-height (get-number-of-lines))))
    
    ;; Returns number of first-level children
    (define (children-num data-list)
          (let ((children-list (assq 'children (cdr data-list))))
            (if children-list
                (length (cdr children-list))
                0)))
    
    ;; Returns total number of children
    (define (children-num-r data-list)
      (cond ((null? data-list) 0)
            ;; No children?
            ((not (assq 'children (cdr data-list))) 1)
            (else (add1 (apply + (map (lambda(elt) (children-num-r elt))
                                      (cdr (assq 'children (cdr data-list)))))))))
    
    ;; Returns total number of non-collapsed children
    (define (children-num-r-nc data-list)
      (cond ((null? data-list) 0)
            ;; No children?
            ((not (assq 'children (cdr data-list))) 1)
            ;; Collapsed?
            ((collapsed? data-list) 1)
            ;; otherwise...
            (else (add1 (apply + (map (lambda(elt) (children-num-r-nc elt))
                                      (cdr (assq 'children (cdr data-list)))))))))
    
    ;; Returns the number of lines in the widget
    (define (get-number-of-lines)
      (if (null? data)
          0
          (children-num-r-nc (car data))))
    
    ;; Quite self-explanatory...
    (define (collapsed? data-list)
      (cadr (assq 'collapsed (cdr data-list))))
    
    ;; Returns true if the element is selected...
    (define (selected? data-list)
      (cadr (assq 'selected (cdr data-list))))
    
    ;; Returns the selected element
    (define (get-selected-element)
      ;; The helper function...
      (letrec ((parse-data-list (lambda (data-list incr offset)
                                  ;; Is the list empty?
                                  (if (not (null? data-list))
                                      (begin
                                        ;; Is the current element selected?
                                        (if (cadr (assq 'selected (cdar data-list)))
                                            ;; It is - then return it, 
                                            ;; if it is not the first element 
                                            ;; (the first element, i.e. the project
                                            ;; element is *never* considered as selected).
                                            (if (= incr 0)  #f (car data-list))
                                            (begin
                                              ;; Otherwise, we need to parse the children...
                                              (or (if (and (assq 'children (cdar data-list))
                                                           (not (collapsed? (car data-list))))
                                                      (parse-data-list (cdr (assq 'children (cdar data-list))) (add1 incr) (add1 offset)) #f)
                                               
                                                  ;; ...and the remaining list.
                                                  (parse-data-list (cdr data-list) 
                                                                   (if (and (assq 'children (cdar data-list)) (collapsed? (car data-list)))
                                                                       (+ 1 incr)
                                                                       (+ 1 incr (sub1 (children-num-r-nc (car data-list)))))
                                                                   offset)))))
                                      ;; There is no selected element...
                                      #f))))
        ;; Then call the helper...
        (parse-data-list data 0 0)))
    
    ; --------------------------------------------------------------------------
    ;                               Swap elements
    ; --------------------------------------------------------------------------
        
    ;; This function returns true iff the element given by 'id' is in the L list.
    (define (in-list? L id)
      ;; If the list is empty, it does not contain the element.
      (cond ((null? L) #f)
            ;; If the first element is the element given by 'id', return true...
            ((eq? (get-property-value (car L) 'id) id) #t)
            ;; Otherwise, search in the cdr.
            (else (in-list? (cdr L) id))))
    
    ;; This function swaps the element given by 'id' with the previous one,
    ;; assuming that the element is in the list and is not the first one.
    (define (swap-element-with-previous L id)
      ;; Did we find the element given by 'id'?
      (cond ((eq? (get-property-value (cadr L) 'id) id)
             ;; Then swap it with the previous one!
             (append (list (cadr L)) (list (car L)) (cddr L)))
             ;; Otherwise, search in the cdr.
            (else (cons (car L) (swap-element-with-previous (cdr L) id)))))
    
    ;; This function swaps the element given by 'id' with the next one,
    ;; assuming that the element is in the list and is not the last one.
    (define (swap-element-with-next L id)
      ;; Did we find the element given by 'id'?
      (cond ((eq? (get-property-value (car L) 'id) id)
             ;; Then swap it with the next one!
             (append (list (cadr L)) (list (car L)) (cddr L)))
            ;; Otherwise, search in the cdr.
            (else (cons (car L) (swap-element-with-next (cdr L) id)))))
    
    ;; This function returns true iff the element given by 'id' is the first
    ;; in the list.
    (define (first-in-list? L id)
      ;; We just check the first element...
      (eq? (get-property-value (car L) 'id) id))
    
    ;; This function returns #t iff the element given by 'id' is the last one
    ;; in the L list, assuming that the element is in the list.
    (define (last-in-list? L id)
      ;; If we found the element, just look whether it is the last one...
      (cond ((eq? (get-property-value (car L) 'id) id) (null? (cdr L)))
            ;; Otherwise, search in the cdr.
            (else (last-in-list? (cdr L) id))))
    
    ;; This function returns #t iff the currently selected element can be swapped (up or down,
    ;; depending on the condition).
    (define (can-swap-selected? id cannot-swap-condition?)
      (let ((can-swap-selected #f))
        ;; This helper procedure will be applied to each line...
        (letrec ((helper (lambda (line)
                           ;; First we get the children list...
                           (let ((children-list (get-children line)))
                             ;; If we found one...
                             ;;  We search whether the element given by 'id' is in the list...
                             (if (and children-list (in-list? children-list id))
                               ;; And then update the result...
                               (set! can-swap-selected 
                                     (not (cannot-swap-condition? children-list id))))))))
          ;; We apply the procedure helper on each element of the data list...
          (apply-procedure-on-data-list data helper)
          ;; And return the result...
          can-swap-selected)))
    
    ;; This function returns #t iff the currently selected element can be swapped down.
    (define (can-swap-down-selected? id)
      ;; The element can be swapped iff it is not the last one in its brother's list.
      (can-swap-selected? id last-in-list?))
    
    ;; This function returns #t iff the currently selected element can be swapped up.
    (define (can-swap-up-selected? id)
      ;; The element can be swapped iff it is not the first one in its brother's list.
      (can-swap-selected? id first-in-list?)) 
    
    ;; This function swaps (up or down depending on the swap function) 
    ;; the currently selected element...
    (define (swap-selected id swap-function)
      ;; This helper procedure will be applied to each line...
      (letrec ((helper (lambda (line)
                       ;; First we get the children list...
                       (let ((children-list (get-children line)))
                         ;; If we found one...
                         (if children-list
                             ;; If the currently selected element is in the list...
                             (if (in-list? children-list id)
                                 ;; Then swap it with the given function! 
                                 ;; And return the updated line...
                                 (set-children line (swap-function children-list id))
                                 ;; Otherwise do not change the line...
                                 line)
                             ;; Otherwise do not change the line...
                             line)))))
        ;; Update the project data...
        (set! data (apply-function-on-data-list data helper))
        ;; And perform miscellaneous updates...
        (update-out)
        (on-paint)
      )
    )
    
    ;; This function swaps up the currently selected element...
    (define (swap-up-selected)
      (let* 
        ((element (get-selected-element))
         (id      (get-property-value element 'id))
        )
        ;; We call swap-selected with the swap-element-with-previous function.
        (swap-selected id swap-element-with-previous)
        ;; We update the preview...
        (let ((parent (project-data-get-parent element)))
          (if parent
            (preview-update-children-order parent)))
      )
      ;; There have been changes since last change...
      (set! no-change-since-last-save #f))
      
    ;; This function swaps down the currently selected element...
    (define (swap-down-selected)
      (let* 
        ((element (get-selected-element))
         (id      (get-property-value element 'id))
        )
        ;; We call swap-selected with the swap-element-with-next function.
        (swap-selected id swap-element-with-next)
        ;; We update the preview...
        (let ((parent (project-data-get-parent element)))
          (if parent
              (preview-update-children-order parent)))
      )
      ;; There have been changes since last change...
      (set! no-change-since-last-save #f))

    ; --------------------------------------------------------------------------
    ;                               Swap elements into
    ; --------------------------------------------------------------------------
    
    (define (can-move-above? obj pparent parent prev first? data-list)
      (cond
        ((null? data-list) #t)
        ((get-property-value (car data-list) 'selected)
         ; the condition is here
         (if (not (or (eq? (caar data-list) 'frame)
                      (and parent (eq? (car parent) 'frame) first?)))
           #t #f)
        )
        (else
         (let ((res #t))
           (if (assq 'children (cdar data-list))
             (set! res (can-move-above? #f ; obj
                                        parent ; pparent
                                        (car data-list) ; parent
                                        prev ;prev
                                        #t   ; first?
                                        (cdr (assq 'children (cdar data-list)))))
           )
           (if res 
               (can-move-above? #f ; obj
                                pparent ; pparent
                                parent  ; parent
                                (car data-list) ;prev
                                #f  ; first?
                                (cdr data-list))
               res
           )
         )
        )
      )
    )
    
    (define (can-move-below? obj pparent parent next last? data-list)
      (cond
        ((null? data-list) #t)
        ((get-property-value (car data-list) 'selected)
         (let ((last? (null? (cdr data-list))))
           ; the condition is here
           (if (or (eq? (caar data-list) 'frame)
                   (and last? (eq? (car parent) 'frame)))
             #f #t)
         )
        )
        (else
         (let* 
           ((res #t)
            (children-list (assq 'children (cdar data-list)))
           )
           (if children-list
             (set! res (can-move-below? #f ; obj
                                        parent ; pparent
                                        (car data-list) ; parent
                                        ;next
                                        (if (null? (cdr data-list))
                                          next
                                          (cadr data-list)
                                        )
                                        #f
                                        (cdr children-list)))
           )
           (if res 
             (can-move-below? #f ; obj
                              pparent ; pparent
                              parent  ; parent
                              next ;next
                              #f
                              (cdr data-list))
             res
           )
         )
        )
      )
    )
    
    ;; This function returns #t iff the currently selected element can be swapped down one.
    (define (can-swap-down-one-selected?)
      (can-move-below? #f #f #f #f #f data)
    )
    
    ;; This function returns #t iff the currently selected element can be swapped up one.
    (define (can-swap-up-one-selected?)
      (can-move-above? #f #f #f #f #f data)
    )
    
    ; Replaces the current parent with a new parent
    (define (set-parent data-list parent)
      (cond ( (null? data-list) 
              data-list)
            ( (symbol? (car data-list)) 
              (cons (car data-list) (set-parent (cdr data-list) parent)))
            ( (eq? (caar data-list) 'parent) 
              (cons (list 'parent parent) (set-parent (cdr data-list) parent)))
            ( else 
              (cons (car data-list) (set-parent (cdr data-list) parent)))))

    ; removes the child with id from the list
    (define (remove-child data-list id)
      (cond ( (null? data-list) 
              data-list)
            ( (eq? (get-property-value (car data-list) 'id) id)
              (cdr data-list)
            )
            ( else 
              (cons (car data-list) (remove-child (cdr data-list) id)))))
    
    ; Inserts an object before an object
    (define (insert-before L id obj)
      (cond ((null? L) L)
            ((eq? (get-property-value (car L) 'id) id)
             ;; Then insert before
             (append (list obj (car L)) (cdr L)))
             ;; Otherwise, search in the cdr.
            (else (cons (car L) (insert-before (cdr L) id obj)))))

    ; Inserts an object after an object
    (define (insert-after L id obj)
      (cond ((null? L) L)
            ((eq? (get-property-value (car L) 'id) id)
             ;; Then insert after
             (append (list (car L) obj) (cdr L)))
             ;; Otherwise, search in the cdr.
            (else (cons (car L) (insert-after (cdr L) id obj)))))

    ; It collects all data for moving a widget upwards in the hierarchy:
    ; obj - the widget data
    ; pparent - parent of the parent of the current widget
    ; parent - parent of the current widget
    ; prev - if the widget is the first in its parent then
    ;        it is the previous widget before the parent
    ;        otherwise it is the previous widget in the parent
    ; first? - whether it was the first widget in the parent
    (define (get-data-for-up obj pparent parent prev first? data-list)
      (cond
        ((null? data-list) #f)
        ((get-property-value (car data-list) 'selected)
         (list (car data-list) pparent parent prev first?))
        (else
         (let ((res #f))
           (if (assq 'children (cdar data-list))
             (set! res (get-data-for-up #f ; obj
                                        parent ; pparent
                                        (get-property-value (car data-list) 'id) ; parent
                                        prev ;prev
                                        #t   ; first?
                                        (cdr (assq 'children (cdar data-list))) ))
           )
           (if res res (get-data-for-up #f ; obj
                                        pparent ; pparent
                                        parent  ; parent
                                        (car data-list) ;prev
                                        #f  ; first?
                                        (cdr data-list) ))
         )
        )
      )
    )
    
    ; It collects all data for moving a widget downwards in the hierarchy:
    ; obj - the widget data
    ; pparent - parent of the parent of the current widget
    ; parent - parent of the current widget
    ; next - if the widget is the last in its parent then
    ;        it is the next widget after the parent
    ;        otherwise it is the next widget in the parent
    ; last? - whether it was the last widget in the parent
    (define (get-data-for-down obj pparent parent next last? data-list)
      (cond
        ((null? data-list) #f) ;if null, then we have not found it
        ; if id matches return data
        ((get-property-value (car data-list) 'selected)
         (let ((last? (null? (cdr data-list))))
           (list (car data-list) pparent parent (if last? next (cadr data-list)) last?)
         )
        )
        (else
         (let ((res #f))
           (if (assq 'children (cdar data-list))
             (set! res (get-data-for-down #f ; obj
                                          parent ; pparent
                                          (get-property-value (car data-list) 'id) ; parent
                                          ;next
                                          (if (null? (cdr data-list))
                                            next
                                            (cadr data-list)
                                          )
                                          #f
                                          (cdr (assq 'children (cdar data-list))) ))
           )
           (if res res (get-data-for-down #f ; obj
                                          pparent ; pparent
                                          parent  ; parent
                                          next ;next
                                          #f
                                          (cdr data-list) ))
         )
        )
      )
    )
    
    ; moves one item up in the hierarchy
    (define (move-up-func data-list)
      ; collect all data
      (let*
        ((info    (get-data-for-up #f #f #f #f #f data-list))
         (obj     (list-ref info 0))
         (pparent (list-ref info 1))
         (parent  (list-ref info 2))
         (prev    (list-ref info 3))
         (first?  (list-ref info 4))
         (id      (get-property-value obj 'id))
         (func-delete
           (lambda (line)
             (let ((children-list (get-children line)))
               (if (and children-list (in-list? children-list id))
                 (set-children line (remove-child children-list id))
                 line
               )
             )
           )
         )
         (func-insert-1
           (lambda (line)
             (if (eq? (get-property-value line 'id) (get-property-value prev 'id))
               ; prev is a container, so simply add to it
               (begin
                 (set! obj (set-parent obj (get-property-value prev 'id)))
                 (set-children line (append (get-children line) (list obj)))
               )
               line)))
         (func-insert-2
           (lambda (line)
             ; first, but prev is not a container then add to pparent before parent
             (if (eq? (get-property-value line 'id) pparent)
               (begin
                 (set! obj (set-parent obj pparent))
                 (set-children line (insert-before (get-children line) parent obj))
               )
               line)))
         (func-insert-3
           (lambda (line)
             ; not first, there is prev but not a container then add before prev
             (if (eq? (get-property-value line 'id) parent)
               (set-children line (insert-before (get-children line) 
                                                 (get-property-value prev 'id) 
                                                 obj))
               line)))
        )
        
;        ; for debugging
;        (display "\nobj: ")(display obj)
;        (display "\npparent: ")(display pparent)
;        (display "\nparent: ")(display parent)
;        (display "\nprev: ")(display prev)
;        (display "\nfirst?: ")(display first?)
        
        ;delete from current position
        (set! data-list (apply-function-on-data-list data-list func-delete))
        
        (cond
          ; current was the first in the list, therefore add current to
          ; parent of the parent of the current, just before the parent
          ( first?
            (apply-function-on-data-list data-list func-insert-2)
          )
          ; the previous widget is a container so add the current to it
          ( (and prev
                 (container? prev)
                 (not (eq? (car prev) 'frame))
            )
            (apply-function-on-data-list data-list func-insert-1)
          )
          ; current was not the first in the list, therefore there is a prev
          ; and add current before prev
          ( else
            (apply-function-on-data-list data-list func-insert-3)
          )
        )        
      )
    )

    ; moves one item down in the hierarchy
    (define (move-down-func data-list)
      ; collect all data
      (let*
        ((info    (get-data-for-down #f #f #f #f #f data-list))
         (obj     (list-ref info 0))
         (pparent (list-ref info 1))
         (parent  (list-ref info 2))
         (next    (list-ref info 3))
         (last?   (list-ref info 4))
         (id      (get-property-value obj 'id))
         (func-delete
           (lambda (line)
             (let ((children-list (get-children line)))
               (if (and children-list (in-list? children-list id))
                 (set-children line (remove-child children-list id))
                 line
               )
             )
           )
         )
         (func-insert-1
           (lambda (line)
             (if (eq? (get-property-value line 'id) (get-property-value next 'id))
               ; next is a container, so simply add to it
               (begin
                 (set! obj (set-parent obj (get-property-value next 'id)))
                 (set-children line (cons obj (get-children line)))
               )
               line)))
         (func-insert-2
           (lambda (line)
             ; last, but next is not a container then add to pparent after parent
             (if (eq? (get-property-value line 'id) pparent)
               (begin
                 (set! obj (set-parent obj pparent))
                 (set-children line (insert-after (get-children line) parent obj))
               )
               line)))
         (func-insert-3
           (lambda (line)
             ; not last, there is next but not a container then add after next
             (if (eq? (get-property-value line 'id) parent)
               (set-children line (insert-after (get-children line)
                                                (get-property-value next 'id) 
                                                obj))
               line)))
        )
        
;        ; for debugging
;        (display "\nobj: ")(display obj)
;        (display "\npparent: ")(display pparent)
;        (display "\nparent: ")(display parent)
;        (display "\nnext: ")(display next)
;        (display "\nlast?: ")(display last?)
        
        ;delete from current position
        (set! data-list (apply-function-on-data-list data-list func-delete))
        
        (cond
          ; current was the last in the list, therefore add current to the
          ; parent of the parent of the current, just after the parent
          ( last?
            (apply-function-on-data-list data-list func-insert-2)
          )
          ; the next widget is a container so add the current to it
          ( (and next
                 (container? next)
                 (not (eq? (car next) 'frame))
            )
            (apply-function-on-data-list data-list func-insert-1)
          )
          ; current was not the last in the list, therefore there is a next
          ; and add current after next
          ( else
            (apply-function-on-data-list data-list func-insert-3)
          )
        )        
      )
    )
        
    ;; This function swaps up the currently selected element...
    (define (swap-up-one-selected)
      (set! data (move-up-func data))
      ;; We update the preview...
      (set! data (preview-update-preview data data))
      (on-paint)
      ;; There have been changes since last change...
      (set! no-change-since-last-save #f)
    )
      
    ;; This function swaps down the currently selected element...
    (define (swap-down-one-selected)
      (set! data (move-down-func data))
      ;; We update the preview...
      (set! data (preview-update-preview data data))
      (on-paint)
      ;; There have been changes since last change...
      (set! no-change-since-last-save #f)
    )
    
    ; --------------------------------------------------------------------------
    ;                               Delete element
    ; --------------------------------------------------------------------------

    ;; Deletes the selected element
    (define (delete-selected)
      
      ;; The helper function which will be passed to apply-function-on-data-list...
      (define (cons-line line)
        (if (and (assq 'id (cdr line))
                 ;; Is the element selected??
                 (eq? (cadr (assq 'id (cdr line))) (get-selected-element-id)))
            'nil
            line))
      
      ;; We provide a different feedback depending whether the element which is about
      ;; to be deleted has children...
      (if (> (children-num (get-selected-element)) 0)
          (if (eq? 'ok (message-box "Warning" 
                                    "This element has children. Deleting it will cause the loss of all its subwidgets. Are you sure you want to continue?" 
                                    hierarchy-frame
                                    '(ok-cancel caution)))
              ;; First, we evaluate the updated data list...
              (begin (let ((next-data (apply-function-on-data-list-improved data cons-line)))
                       ;; ...then, we delete the element in the preview...
                       (preview-delete-element (get-selected-element) next-data)
                       ;; ...and then in the data list
                       (set! data next-data)
                       ;; And then perform needed updates...
                       (send properties-widget empty)
                       (set! no-change-since-last-save #f)
                       (update-out)
                       (on-paint))))
          
          ;; First, delete the element in the preview
          (begin (let ((next-data (apply-function-on-data-list-improved data cons-line)))
                   ;; ...then, we delete the element in the preview...
                   (preview-delete-element (get-selected-element) next-data)
                   ;; ...and then in the data list
                   (set! data next-data)
                   ;; And then perform needed updates...
                   (send properties-widget empty)
                   (set! no-change-since-last-save #f)
                   (update-out)
                   (on-paint)))))
    
    ;; Returns the selected element id
    (define (get-selected-element-id)
      (cadr (assq 'id (cdr (get-selected-element)))))
    
    ;; This function dispatches the action to the right line
    (define (action-line num action)
      
      ;; This sub-routine sets a children list to an element
      (define (set-children data-list children)
        (cond ((null? data-list) data-list)
              ((symbol? (car data-list)) (cons (car data-list) (set-children (cdr data-list) children)))
              ((eq? (caar data-list) 'children) (cons (cons 'children children) (set-children (cdr data-list) children)))
              (else (cons (car data-list) (set-children (cdr data-list) children)))))
      
      ;; This sub-routine selects/deselects an element
      (define (select-element data-list selected)
        (cond ((null? data-list) data-list)
              ((symbol? (car data-list)) (cons (car data-list) (select-element (cdr data-list) selected)))
              ((eq? (caar data-list) 'selected) (cons (list 'selected selected) (select-element (cdr data-list) selected)))
              (else (cons (car data-list) (select-element (cdr data-list) selected)))))  
      
      ;; This sub-routine collapses/uncollapses an element
      (define (collapse-element data-list change)
        (cond ((null? data-list) data-list)
              ((symbol? (car data-list)) (cons (car data-list) (collapse-element (cdr data-list) change)))
              ((eq? (caar data-list) 'collapsed) 
               (cons (list 'collapsed (if change (not (cadar data-list)) (cadar data-list))) 
                     (collapse-element (cdr data-list) change)))
              (else (cons (car data-list) (collapse-element (cdr data-list) change)))))   
      
      ;; This function iterates on each list element.
      (define (helper data-list incr offset)
        (if (null? data-list)
            data-list
            (begin
              (let* ((mod-current-line        
                      ;; Is it the right line?
                      (if (eq? action 'select)
                          (if (= incr num)
                              (select-element (car data-list) #t)
                              (select-element (car data-list) #f))
                          (if (= incr num)
                              (collapse-element (car data-list) #t)
                              (collapse-element (car data-list) #f))))
                     
                     (mod-current-children
                      ;; Do the current element have children?
                      (if (and (assq 'children (cdar data-list))
                               (not (collapsed? (car data-list))))
                          (helper (cdr (assq 'children (cdar data-list))) (add1 incr) (add1 offset))
                          #f)))
                
                ;; We build the result and call the second recursive call
                (cons (if mod-current-children
                          (set-children mod-current-line mod-current-children)
                          mod-current-line)
                      (helper (cdr data-list) 
                              (if (and (assq 'children (cdar data-list))
                                       (collapsed? (car data-list)))
                                  (+ 1 incr)
                                  (+ 1 incr (sub1 (children-num-r-nc (car data-list)))))
                              offset))))))
      
      ;; The real function body
      (set! data (helper data 0 0))
      (on-paint))
    
    ;; This function updates the enabled/disabled buttons status.
    (define (update-buttons)
      ;; We are going to check whether the currently selected line is a window
      ;; and whether it is shown in order to activate or deactivate the show/hide buttons
      (let ((selected-element (get-selected-element)))
        (if selected-element
            (let ((id (get-property-value selected-element 'id)))
              (if (eq? (car selected-element) 'frame)
                  ;; Ok this is a frame. Is it shown?
                  (if (send (cadr (assq 'pointer (cdr selected-element))) is-shown?)
                      (begin
                        (send hierarchy-show-button enable #f)
                        (send hierarchy-hide-button enable #t))
                      (begin 
                        (send hierarchy-show-button enable #t)
                        (send hierarchy-hide-button enable #f)))
                  ;; This is not a frame. Therefore disable both buttons.
                  (begin
                    (send hierarchy-show-button enable #f)
                    (send hierarchy-hide-button enable #f)))
              (if (eq? (car selected-element) 'project)
                  (send hierarchy-delete-button enable #f)
                  (send hierarchy-delete-button enable #t))
              (if (can-swap-down-one-selected?)
                  (send hierarchy-down-one-button enable #t)
                  (send hierarchy-down-one-button enable #f))
              (if (can-swap-up-one-selected?)
                  (send hierarchy-up-one-button enable #t)
                  (send hierarchy-up-one-button enable #f))
              (if (can-swap-down-selected? id)
                  (send hierarchy-down-button enable #t)
                  (send hierarchy-down-button enable #f))
              (if (can-swap-up-selected? id)
                  (send hierarchy-up-button enable #t)
                  (send hierarchy-up-button enable #f)))
            (begin
              (send hierarchy-up-one-button enable #f)
              (send hierarchy-down-one-button enable #f)
              (send hierarchy-up-button enable #f)
              (send hierarchy-down-button enable #f)
              (send hierarchy-delete-button enable #f)
              (send hierarchy-show-button enable #f)
              (send hierarchy-hide-button enable #f)))))
    
    
    ;; This function updates the external environment with local data.
    (define (update-out)
      (set! project-data-global data))
    
    ;; This function updates the bitmap in which drawings are performed before
    ;; appearing on the screen.
    (define (update-bitmap)
      ;; We will need the virtual size...
      (let-values (((width height) (send this get-virtual-size)))
        ;; We create a bitmap, with the *real* width and height (at least, if they are
        ;; greater than the virtual width and height...), in order to be able to scroll very quickly...
        (set! bitmap (make-object bitmap% (max (get-real-width) width 1) (max (get-real-height) height 1) #f)))
      ;; And then, we update the associated bitmap-dc...
      (set! dc (instantiate bitmap-dc% (bitmap))))

    ;; This function updates the scroll bars properties...
    (define (update-scroll-bars)
      ;; We need to get the virtual size...
      (let-values (((width height) (send this get-virtual-size)))
        ;; Then, we update the scroll range for both the horizontal and the vertical scroll
        ;; bars. The right range is the full real height (width, respectively) minus the
        ;; virtual height (resp. width).
        (send this set-scroll-range 'horizontal (max 0 (- (get-real-width) width)))
        (send this set-scroll-range 'vertical (max 0 (- (get-real-height) height)))
        ;; If the window can contain the whole height, no y offset is needed.
        (if (= 0 (send this get-scroll-range 'vertical)) (set! y-offset 0) (on-paint))
        ;; If the window can contain the whole width, no x offset is needed.
        (if (= 0 (send this get-scroll-range 'horizontal)) (set! x-offset 0) (on-paint))
        ;; This is the number of scroll steps - vertically we want to scroll not exactly one
        ;; page, but one page minus one line (as it is always implemented).
        (send this set-scroll-page 'horizontal (max 1 width))
        (send this set-scroll-page 'vertical (max 1 (- height k-line-height)))))
    
    ;; Overriden methods
    (override on-scroll on-size on-paint on-event on-char)
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-scroll                                 |   
    ;; +----------------------------------------------------------------------+
    (define (on-scroll scroll-event)
      ;; What we are going to do depends on the event direction...
      (case (send scroll-event get-direction)
        ;; Is it vertical?
        ((vertical) 
         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;;
           ;; Notice : I am forced to use eq?, because I have noticed a bug under MacOS X : if the event is a thumb event,
           ;; DrScheme crashes if we try to evaluate the result given by get-event-type! It looks to return an invalid
           ;; pointer, at a system adress, maybe 0. Therefore I just compare pointers.
           ;;
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions (to force scrolling more than 1 step [1 line, actually])...
                    (send this set-scroll-pos 'vertical (min (+ (send this get-scroll-pos 'vertical) k-line-height -1) 
                                                             (send this get-scroll-range 'vertical)))
                    ;; ...and the vertical offset.
                    (set! y-offset (min (+ y-offset k-line-height) (send this get-scroll-range 'vertical)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'vertical (max (- (send this get-scroll-pos 'vertical) k-line-height -1) 0))
                  ;; ...and the vertical offset.
                  (set! y-offset (max (- y-offset k-line-height) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, just to update the vertical offset...
                 (else (set! y-offset (send scroll-event get-position))))))
        
        ;; Is it horizontal?
        ((horizontal) 
         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions (to force scrolling more than 1 step... [1 "horizontal scroll step", actually])...
                    (send this set-scroll-pos 'horizontal (min (+ (send this get-scroll-pos 'horizontal) k-horizontal-scroll-step -1) 
                                                               (send this get-scroll-range 'horizontal)))
                    ;; ...and the horizontal offset.
                    (set! x-offset (min (+ x-offset k-horizontal-scroll-step) (send this get-scroll-range 'horizontal)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'horizontal (max (- (send this get-scroll-pos 'horizontal) k-horizontal-scroll-step -1) 0))
                  ;; ...and the horizontal offset.
                  (set! x-offset (max (- x-offset k-horizontal-scroll-step) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, just to update the horizontal offset...
                 (else (set! x-offset (send scroll-event get-position)))))))
      
      ;; And then, we directly draw the bitmap, which has already been drawed during the last
      ;; call of (on-paint), at the right position - which is *much* faster than calling on-paint
      ;; directly and allows a smooth scrolling.
      (send (send this get-dc) draw-bitmap bitmap (- x-offset) (- y-offset)))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-size                                   |   
    ;; +----------------------------------------------------------------------+
    (define (on-size width height)
      ;; We just update the bitmap...
      (update-bitmap)
      ;; ...and the scroll bars.
      (update-scroll-bars))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-paint                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-paint)
      (let* ((width              (send this get-width))
             (height             (send this get-height)))
        
        ;; Data list parser
        (define (parse-data-list data-list incr offset)
          (if (not (null? data-list))
              (begin
                ;; Text foreground is black...
                (send dc set-text-foreground k-black)
                
                ;; Set the right font...
                (cond ((memq (caar data-list) '(frame project)) (send dc set-font k-bold-font))
                      ((container? (car data-list)) (send dc set-font k-italic-font))
                      (else (send dc set-font k-normal-font)))
                
                ;; The project name is on a gray background.
                (if (zero? offset)
                    (begin
                      (send dc set-brush k-gray-brush)
                      (send dc draw-rectangle 0 
                            (* incr k-line-height) (send bitmap get-width) k-line-height)))
                
                ;; Let's draw the links betwixt the +/- boxes.
                ;; First we choose the color...
                (send dc set-brush k-dkgray-brush)
                
                ;; We only draw them if the current element has children.
                (if (not (zero? (children-num (car data-list))))
                    ;; The lines function parses the children in order to find
                    ;; which must and which must not be linked with the current element.
                    (letrec ((lines (lambda (n list)
                                      (cond
                                        ((null? list) (void))
                                        ((> (children-num (car list)) 0)
                                         (trace n)
                                         (if (collapsed? (car list))
                                             (lines (+ n 1) (cdr list))
                                             (lines (+ n (children-num-r-nc (car list))) (cdr list))))
                                        (else (lines (+ n 1) (cdr list))))))
                             ;; The trace function actually draws the links.
                             (trace (lambda (n)
                                      (begin
                                        (send dc draw-rectangle (+ (* 20 offset) 10)
                                              (+ (* (+ n incr) k-line-height) 8) 16 1)
                                        (send dc draw-rectangle (+ (* 20 offset) 9) 
                                              (+ (* incr k-line-height) 4)
                                              1 (+ (* n k-line-height) 5))))))
                      ;; Then we call the lines function, but only if the current
                      ;; element is not collapsed.
                      (if (not (collapsed? (car data-list)))
                          (lines 1 (cdr (assq 'children (cdar data-list)))))))
                (if (assq (caar data-list) k-icon-list)
                ;; We draw the right icon...
                (send dc draw-bitmap 
                      ;; We look for the right icon in the a-list...
                      (cdr (assq (caar data-list) k-icon-list)) 
                      (+ (* 20 offset) k-left-margin)
                      (+ (* k-line-height incr) 1) 
                      'solid 
                      (make-object color% "black")
                      (if (eq? (caar data-list) 'project)
                          k-project-mask-bitmap
                          #f)))
                
                ;; Draw the +/- icons
                (if (not (zero? (children-num (car data-list))))
                    ;; If the current element is collapsed...
                    (if (not (collapsed? (car data-list)))
                        ;; ...we draw a +...
                        (send dc draw-bitmap 
                              k-less-bitmap
                              (+ (* 20 offset) 5)
                              (+ (* k-line-height incr) 4) 
                              'solid (make-object color% "white") #f)
                        ;; ...we draw a - otherwise.
                        (send dc draw-bitmap 
                              k-plus-bitmap
                              (+ (* 20 offset) 5)
                              (+ (* k-line-height incr) 4) 
                              'solid (make-object color% "white") #f))) 
                
                (send dc set-text-foreground k-black)
                
                ;; Display the caption...
                (send dc draw-text (if (assq 'label (cdar data-list))
                                       (let ((label (cadr (assq 'label (cdar data-list)))))
                                         (if (and label (not (string=? label ""))) label "(no label)"))
                                       (string-append "<" (symbol->string (caar data-list)) ">"))
                      (+ (* 20 offset) k-left-margin 20) 
                      (+ (* k-line-height incr) k-top-margin))
                
                (let-values (((width height distance extra)
                              (send dc get-text-extent 
                                    (if (assq 'label (cdar data-list))
                                      (let ((label (cadr (assq 'label (cdar data-list)))))
                                        (if (and label (not (string=? label ""))) label "(no label)"))
                                      (string-append "<" (symbol->string (caar data-list)) ">")))))
                  (send dc set-text-foreground k-gray)
                  
                  ;; Display the id...
                  (send dc set-font k-italic-smoothed-font)
                  (send dc draw-text 
                        (let ((id-v (assq 'id (cdar data-list))))
                          (if id-v
                            (string-append " [" 
                                           (symbol->string (cadr (assq 'id (cdar data-list))))
                                           "]")
                            ""))
                        (+ (* 20 offset) k-left-margin 20 width) 
                        (+ (* k-line-height incr) k-top-margin))
                  
                  ;; Is the current element selected?
                  (if (selected? (car data-list))
                      ;; Then draw the selection!
                      (begin
                        (send dc set-brush k-active-brush)
                        (send dc draw-rectangle 
                              (+ (* 20 offset) k-left-margin 19) 
                              (+ (* k-line-height incr) distance -1)
                              (+ width 2)
                              (- k-line-height 2)))))
                
                
                ;; Do the current element have children?
                (if (and (assq 'children (cdar data-list))
                         (not (collapsed? (car data-list))))
                    ;; Then show them!
                    (parse-data-list (cdr (assq 'children (cdar data-list))) 
                                     (add1 incr) 
                                     (add1 offset)))
                
                ;; Continue with the rest of the list...
                (parse-data-list (cdr data-list) 
                                 ;; Next line is one plus current line plus number of children
                                 ;; if the element is not collapsed, the current line plus 1
                                 ;; otherwise
                                 (if (and (assq 'children (cdar data-list)) 
                                          (collapsed? (car data-list)))
                                     (+ 1 incr)
                                     (+ 1 incr (sub1 (children-num-r-nc (car data-list)))))
                                 ;; No offset, we are at the same hierarchy level
                                 offset))))
        
        (send dc set-pen k-invisible-pen) ;; Invisible pen
        (send dc set-brush k-white-brush) ;; White brush

        ;; We clear the bitmap...
        (send dc clear)
        
        ;; ...and then parse the list
        (parse-data-list data 0 0)
        
        ;; We drawed the whole output into a bitmap in order to avoid flickering,
        ;; therefore we have to draw the bitmap.
        (if (send bitmap ok?)
            (send (send this get-dc) draw-bitmap bitmap (- x-offset) (- y-offset)))))
    
    
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-event                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-event event)
      ;; The widget gets the focus
      (send this focus)
      
      ;; On-event body starts here
      (let ((x    (+ (send event get-x) x-offset))
            (y    (+ (send event get-y) y-offset)))
        (case (send event get-event-type)
          ;; Mouse-down event with the left mouse button
          ((left-down) 
           (letrec 
             ( (parse-data-list 
                 (lambda (data-list incr offset)
                   (if (not (null? data-list))
                     (begin
                       (let-values 
                         (((width height distance extra)
                           (send dc get-text-extent 
                                 ;; The label text...
                                 (if (assq 'label (cdar data-list))
                                   (let ((label (cadr (assq 'label (cdar data-list)))))
                                     (if (and label (not (string=? label ""))) label "(no label)"))
                                   (string-append "<" (symbol->string (caar data-list)) ">"))
                                 ;; And its font...
                                 (cond
                                   ((memq (caar data-list) '(frame project)) k-bold-font)
                                   ((container? (car data-list)) k-italic-font)
                                   (else k-normal-font)))))
                         
                         ;; Is the event a field area?
                         (if (and (>= x (+ (* 20 offset) k-left-margin ))
                                  (<= x (+ (* 20 offset) k-left-margin width 35))
                                  (>= y (+ (* k-line-height incr) distance -1))
                                  (<= y (+ (* k-line-height incr) distance 1 height)))
                           (action-line incr 'select))
                                               
                         ;;Is the event in a +/- area?
                         (if (and (>= x (+ (* 20 offset) k-left-margin -16))
                                  (<= x (+ (* 20 offset) k-left-margin -7))
                                  (>= y (+ (* k-line-height incr) 4))
                                  (<= y (+ (* k-line-height incr) 13)))
                           (action-line incr 'collapse))
                                               
                         (if (and (assq 'children (cdar data-list))
                                  (not (collapsed? (car data-list))))
                           (parse-data-list (cdr (assq 'children (cdar data-list)))
                                            (add1 incr) (add1 offset)))
                         
                         ;; Parse the list
                         (parse-data-list (cdr data-list) 
                                          (if (and (assq 'children (cdar data-list))
                                                   (collapsed? (car data-list)))
                                            (+ 1 incr)
                                            (+ 1 incr (sub1 (children-num-r-nc (car data-list)))))
                                                                offset)))))))
             
             ;; Parse the whole list
             (parse-data-list data 0 0)
             
             ;; Update the project data
             (set! project-data-global data)
             
             ;; Update the property widget
             (send properties-widget update-in)
             
             ;; Update the buttons
             (send properties-widget update-buttons)
             (update-buttons))))))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-char                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-char event)
      (case (send event get-key-code)
        
        ;; Wheel up
        ((wheel-up)
         (on-scroll (new scroll-event% (event-type 'line-up) (direction 'vertical))))
        
        ;; Wheel down
        ((wheel-down)
         (on-scroll (new scroll-event% (event-type 'line-down) (direction 'vertical))))))
            

    ;; Call the initialization function from the parent class...
    (super-new (style '(vscroll hscroll border)))
    ;; ...ok, now we can call 'this'.
    
    ;; We initialize the scrollbars - maybe isn't it needed?
    ;; For compatibility reasons, we will not delete the following line.
    (send this init-manual-scrollbars #f #f 1 1 0 0)
    
    ;; We then need to call on-size so that the bitmap and
    ;; drawing context are initialised
    (on-size (send (send this get-parent) get-width) (send (send this get-parent) get-height))))

